@@ -95,6 +95,57 @@ use crate::{
 #[derive(Debug, Parser)]
 #[clap(about = "An implementation of the Stateless OpenPGP Command Line Interface")]
 #[clap(disable_version_flag(true))]
+struct Cli {
+    /// Emit verbose output for debugging.
+    #[clap(long = "debug", global = true)]
+    debug: bool,
+
+    #[clap(subcommand)]
+    operation: Operation,
+}
+
+/// Variant of the SOP specification.
+pub enum Variant {
+    /// The full functionality.
+    Full,
+
+    /// The verification subset.
+    Verification,
+}
+
+impl Cli {
+    fn with_version<'v, V>(v: &V, variant: Variant)
+                       -> Result<clap::Command>
+    where
+        V: crate::ops::Version<'v> + ?Sized
+    {
+        // Do a little dance to supply name and version to clap.
+        let (name, version, about) = {
+            let v = v.frontend()?;
+            (
+                v.name.clone(),
+                v.version.clone(),
+                format!("An implementation of the {}\
+                         Stateless OpenPGP Command \
+                         Line Interface using {} {}",
+                        match variant {
+                            Variant::Full => "",
+                            Variant::Verification =>
+                                "verification-only subset of the ",
+                        },
+                        v.name, v.version),
+            )
+        };
+
+        Ok(Cli::command()
+           .name(Box::leak(name.into_boxed_str()) as &str)
+           .about(Box::leak(about.into_boxed_str()) as &str)
+           .version(Box::leak(version.into_boxed_str()) as &str))
+    }
+}
+
+
+#[derive(Debug, Parser)]
 enum Operation {
     /// Prints version information.
     ///
@@ -198,13 +249,101 @@ enum Operation {
     },
 
     /// Extracts a Certificate from a Secret Key.
-    #[clap(display_order = 300)]
+    #[clap(display_order = 230)]
     #[cfg(feature = "cli")]
     ExtractCert {
         /// Don't ASCII-armor output.
         #[clap(long)]
         no_armor: bool,
     },
+
+    /// Keep a Secret Key Up-To-Date.
+    #[clap(display_order = 240)]
+    #[cfg(feature = "cli")]
+    UpdateKey {
+        /// Don't ASCII-armor output.
+        #[clap(long)]
+        no_armor: bool,
+
+        /// Don't make the updated key encryption-capable if it isn't
+        /// already.
+        #[clap(long)]
+        signing_only: bool,
+
+        /// Don't advertise support for capabilities that aren't
+        /// already advertised by the key.
+        #[clap(long)]
+        no_added_capabilities: bool,
+
+        /// Unlock the keys with these passwords.
+        #[clap(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
+
+        /// Merge updates into the key.
+        #[clap(long, number_of_values = 1)]
+        merge_certs: Vec<String>,
+    },
+
+    /// Merge OpenPGP Certificates.
+    #[clap(display_order = 250)]
+    #[cfg(feature = "cli")]
+    MergeCerts {
+        /// Don't ASCII-armor output.
+        #[clap(long)]
+        no_armor: bool,
+
+        /// Merge updates into the certs.
+        certs: Vec<String>,
+    },
+
+    /// Certify OpenPGP Certificate User IDs.
+    #[clap(display_order = 300)]
+    #[cfg(feature = "cli")]
+    CertifyUserid {
+        /// Don't ASCII-armor output.
+        #[clap(long)]
+        no_armor: bool,
+
+        /// Certify the specified user IDs.
+        #[clap(long, required = true, number_of_values = 1)]
+        userid: Vec<String>,
+
+        /// Unlock the keys with these passwords.
+        #[clap(long, number_of_values = 1)]
+        with_key_password: Vec<String>,
+
+        /// Don't require self-signatures on the user IDs to be
+        /// certified.
+        #[clap(long)]
+        no_require_self_sig: bool,
+
+        /// Create certifications using these keys.
+        keys: Vec<String>,
+    },
+
+    /// Validate a User ID in an OpenPGP Certificate.
+    #[clap(display_order = 320)]
+    #[cfg(feature = "cli")]
+    ValidateUserid {
+        /// Treat USERID as an e-mail address, and matched only
+        /// against the e-mail address part of each correctly bound
+        /// User ID.
+        #[clap(long)]
+        addr_spec_only: bool,
+
+        /// Evaluate the validity of the User ID at the specified
+        /// time, not at the current time.
+        #[clap(long)]
+        validate_at: Option<Timestamp>,
+
+        /// The User ID to validate.
+        userid: String,
+
+        /// Authority OpenPGP certificates, i.e. trust roots.
+        #[clap(required = true)]
+        certs: Vec<String>,
+    },
+
     /// Creates Detached Signatures.
     #[clap(display_order = 400)]
     #[cfg(feature = "cli")]
@@ -373,17 +512,43 @@ where B: AsRef<str>,
     let out_path = out_path.as_ref();
     std::fs::create_dir_all(&out_path)?;
 
-    let mut clap = Operation::command();
+    let mut clap = Cli::command();
     for shell in &[Shell::Bash, Shell::Fish, Shell::Zsh, Shell::PowerShell,
                    Shell::Elvish] {
-        let path = generate_to(*shell, &mut clap, binary_name, out_path)?;
-        println!("cargo:warning=completion file is generated: {:?}", path);
+        generate_to(*shell, &mut clap, binary_name, out_path)?;
     }
+
+    println!("cargo:warning=shell completions written to {}",
+             out_path.display());
+    Ok(())
+}
+
+/// Generates man pages.
+pub fn write_man_pages<A, P, V>(variant: Variant,
+                                version: &V,
+                                author: A,
+                                out_path: P)
+                                -> Result<()>
+where
+    V: for <'a> crate::ops::Version<'a>,
+    A: ToString,
+    P: AsRef<Path>,
+{
+    let out_path = out_path.as_ref();
+    std::fs::create_dir_all(&out_path)?;
+
+    let cmd = Cli::with_version(version, variant)?
+        .author(Box::leak(author.to_string().into_boxed_str()) as &str);
+    clap_mangen::generate_to(cmd, out_path)?;
+
+    println!("cargo:warning=man pages written to {}",
+             out_path.display());
     Ok(())
 }
 
 /// Implements the SOP command-line interface.
-pub fn main<'s, S, C, K, Sigs>(sop: &'s S) -> !
+pub fn main<'s, S, C, K, Sigs>(sop: &'s mut S, variant: Variant)
+                               -> !
 where
     S: SOP<'s, Certs = C, Keys = K, Sigs = Sigs>,
     C: Load<'s, S> + Save,
@@ -392,7 +557,7 @@ where
 {
     use std::process::exit;
 
-    match real_main::<S, C, K, Sigs>(sop) {
+    match real_main::<S, C, K, Sigs>(sop, variant) {
         Ok(()) => exit(0),
         Err(e) => {
             print_error_chain(&e);
@@ -413,24 +578,15 @@ where
     }
 }
 
-fn real_main<'s, S, C, K, Sigs>(sop: &'s S) -> Result<()>
+fn real_main<'s, S, C, K, Sigs>(sop: &'s mut S, variant: Variant)
+                                -> Result<()>
 where
     S: SOP<'s, Certs = C, Keys = K, Sigs = Sigs>,
     C: Load<'s, S> + Save,
     K: Load<'s, S> + Save,
     Sigs: Load<'s, S> + Save,
 {
-    // Do a little dance to supply name and version to clap.
-    let v = sop.version()?.frontend()?;
-    let name = format!("SOP {}", v.name);
-    let version = v.version.clone();
-    let about = format!("An implementation of the Stateless OpenPGP Command \
-                        Line Interface using {} {}",
-                        v.name, v.version);
-    let app = Operation::command()
-        .name(Box::leak(name.into_boxed_str()) as &str)
-        .about(Box::leak(about.into_boxed_str()) as &str)
-        .version(Box::leak(version.into_boxed_str()) as &str);
+    let app = Cli::with_version(sop.version()?.as_ref(), variant)?;
 
     let matches = match app.clone().try_get_matches() {
         Ok(v) => v,
@@ -448,7 +604,10 @@ where
         },
     };
 
-    match Operation::from_arg_matches(&matches)? {
+    let cli = Cli::from_arg_matches(&matches)?;
+    sop.debug(cli.debug);
+
+    match cli.operation {
         Operation::Version { backend, extended, sop_spec, sopv } => {
             let version = sop.version()?;
             match (backend, extended, sop_spec, sopv) {
@@ -510,7 +669,7 @@ where
                 op = op.userid(&u);
             }
             op.generate()?
-                .to_writer(! no_armor, &mut io::stdout())?;
+                .to_stdout(! no_armor)?;
         },
 
         #[cfg(feature = "cli")]
@@ -526,8 +685,8 @@ where
             for p in old_key_password {
                 op = op.old_key_password(get_password(p)?)?;
             }
-            op.keys(&K::from_reader(sop, &mut io::stdin())?)?
-                .to_writer(! no_armor, &mut io::stdout())?;
+            op.keys(&K::from_stdin(sop)?)?
+                .to_stdout(! no_armor)?;
         }
 
         #[cfg(feature = "cli")]
@@ -536,15 +695,89 @@ where
             for p in with_key_password {
                 op = op.with_key_password(get_password(p)?)?;
             }
-            op.keys(&K::from_reader(sop, &mut io::stdin())?)?
-                .to_writer(! no_armor, &mut io::stdout())?;
+            op.keys(&K::from_stdin(sop)?)?
+                .to_stdout(! no_armor)?;
         }
 
         #[cfg(feature = "cli")]
         Operation::ExtractCert { no_armor, } => {
             sop.extract_cert()?
-                .keys(&K::from_reader(sop, &mut io::stdin())?)?
-                .to_writer(! no_armor, &mut io::stdout())?;
+                .keys(&K::from_stdin(sop)?)?
+                .to_stdout(! no_armor)?;
+        },
+
+        #[cfg(feature = "cli")]
+        Operation::UpdateKey {
+            no_armor, signing_only, no_added_capabilities,
+            with_key_password, merge_certs,
+        } => {
+            let mut op = sop.update_key()?;
+            if signing_only {
+                op = op.signing_only();
+            }
+            if no_added_capabilities {
+                op = op.no_added_capabilities();
+            }
+            for p in with_key_password {
+                op = op.with_key_password(get_password(p)?)?;
+            }
+            for (name, mut stream) in load_files(merge_certs)? {
+                op = op.merge_updates(&C::from_reader(sop, &mut stream, Some(name))?)?;
+            }
+            let keys = op.update(&K::from_stdin(sop)?)?;
+            keys.to_stdout(! no_armor)?;
+        },
+
+        #[cfg(feature = "cli")]
+        Operation::MergeCerts { no_armor, certs, } => {
+            let mut op = sop.merge_certs()?;
+            for (name, mut stream) in load_files(certs)? {
+                op = op.merge_updates(&C::from_reader(sop, &mut stream, Some(name))?)?;
+            }
+            let certs = op.merge(&C::from_stdin(sop)?)?;
+            certs.to_stdout(! no_armor)?;
+        },
+
+        #[cfg(feature = "cli")]
+        Operation::CertifyUserid {
+            no_armor, userid, with_key_password, no_require_self_sig, keys,
+        } => {
+            let mut op = sop.certify_userid()?;
+            if no_require_self_sig {
+                op = op.no_require_self_sig();
+            }
+            for u in userid {
+                op = op.userid(u);
+            }
+            for p in with_key_password {
+                op = op.with_key_password(get_password(p)?)?;
+            }
+            for (name, mut stream) in load_files(keys)? {
+                op = op.keys(&K::from_reader(sop, &mut stream, Some(name))?)?;
+            }
+            let certs = op.certify(&C::from_stdin(sop)?)?;
+            certs.to_stdout(! no_armor)?;
+        },
+
+        #[cfg(feature = "cli")]
+        Operation::ValidateUserid {
+            addr_spec_only, validate_at, userid, certs,
+        } => {
+            let mut op = sop.validate_userid()?;
+            if let Some(t) = validate_at {
+                op = op.validate_at(t.into())?;
+            }
+            for (name, mut stream) in load_files(certs)? {
+                op = op.trust_roots(&C::from_reader(sop, &mut stream, Some(name))?)?;
+            }
+
+            op = op.target_certs(&C::from_stdin(sop)?)?;
+
+            if addr_spec_only {
+                op.email(&userid)?;
+            } else {
+                op.userid(&userid)?;
+            }
         },
 
         #[cfg(feature = "cli")]
@@ -559,15 +792,15 @@ where
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut stream in load_files(keys)? {
-                op = op.keys(&K::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(keys)? {
+                op = op.keys(&K::from_reader(sop, &mut stream, Some(name))?)?;
             }
             let (micalg, sigs) = op.data(&mut io::stdin())?;
             if let Some(path) = micalg_out {
                 let mut sink = create_file(path)?;
                 write!(sink, "{}", micalg)?;
             }
-            sigs.to_writer(! no_armor, &mut io::stdout())?;
+            sigs.to_stdout(! no_armor)?;
         },
 
         #[cfg(any(feature = "cli", feature = "cliv"))]
@@ -579,11 +812,12 @@ where
             if let Some(t) = not_after {
                 op = op.not_after(t.into());
             }
-            for mut stream in load_files(certs)? {
-                op = op.certs(&C::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(certs)? {
+                op = op.certs(&C::from_reader(sop, &mut stream, Some(name))?)?;
             }
+            let (name, mut stream) = load_file(signatures)?;
             let verifications =
-                op.signatures(&Sigs::from_reader(sop, &mut load_file(signatures)?)?)?
+                op.signatures(&Sigs::from_reader(sop, &mut stream, Some(name))?)?
                 .data(&mut io::stdin())?;
 
             for v in verifications {
@@ -620,14 +854,14 @@ where
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut stream in load_files(sign_with)? {
-                op = op.sign_with_keys(&K::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(sign_with)? {
+                op = op.sign_with_keys(&K::from_reader(sop, &mut stream, Some(name))?)?;
             }
             for pw in with_password.into_iter().map(get_password) {
                 op = op.with_password(pw?)?;
             }
-            for mut stream in load_files(certs)? {
-                op = op.with_certs(&C::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(certs)? {
+                op = op.with_certs(&C::from_reader(sop, &mut stream, Some(name))?)?;
             }
             let session_key =
                 op.plaintext(&mut io::stdin())?.to_writer(&mut io::stdout())?;
@@ -675,7 +909,7 @@ where
 
             let mut op = sop.decrypt()?;
 
-            for mut stream in load_files(with_session_key)? {
+            for (_name, mut stream) in load_files(with_session_key)? {
                 let mut sk = String::new();
                 stream.read_to_string(&mut sk)?;
                 op = op.with_session_key(sk.parse()?)?;
@@ -689,12 +923,12 @@ where
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
 
-            for mut stream in load_files(keys)? {
-                op = op.with_keys(&K::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(keys)? {
+                op = op.with_keys(&K::from_reader(sop, &mut stream, Some(name))?)?;
             }
 
-            for mut stream in load_files(verify_with)? {
-                op = op.verify_with_certs(&C::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(verify_with)? {
+                op = op.verify_with_certs(&C::from_reader(sop, &mut stream, Some(name))?)?;
             }
 
             if let Some(t) = verify_not_before {
@@ -765,8 +999,8 @@ where
                 } else {
                     Box::new(io::sink())
                 };
-            for mut stream in load_files(certs)? {
-                op = op.certs(&C::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(certs)? {
+                op = op.certs(&C::from_reader(sop, &mut stream, Some(name))?)?;
             }
             let verifications =
                 op.message(&mut io::stdin())?.to_writer(&mut io::stdout())?;
@@ -785,8 +1019,8 @@ where
             for p in with_key_password {
                 op = op.with_key_password(get_password_unchecked(p)?)?;
             }
-            for mut stream in load_files(keys)? {
-                op = op.keys(&K::from_reader(sop, &mut stream)?)?;
+            for (name, mut stream) in load_files(keys)? {
+                op = op.keys(&K::from_reader(sop, &mut stream, Some(name))?)?;
             }
             op.data(&mut io::stdin())?.to_writer(&mut io::stdout())?;
         },
@@ -805,8 +1039,9 @@ fn is_special_designator<S: AsRef<str>>(file: S) -> bool {
 
 /// Loads the given (special) file.
 fn load_file<S: AsRef<str>>(file: S)
-                            -> Result<Box<dyn io::Read + Send + Sync>> {
+                            -> Result<(String, Box<dyn io::Read + Send + Sync>)> {
     let f = file.as_ref();
+    let file_name = f.into();
 
     if is_special_designator(f) {
         if Path::new(f).exists() {
@@ -826,7 +1061,7 @@ fn load_file<S: AsRef<str>>(file: S)
                 let f = unsafe {
                     std::fs::File::from_raw_fd(fd)
                 };
-                return Ok(Box::new(f));
+                return Ok((file_name, Box::new(f)));
             }
 
             if f.starts_with("@ENV:") {
@@ -837,7 +1072,8 @@ fn load_file<S: AsRef<str>>(file: S)
                     .ok_or(Error::UnsupportedSpecialPrefix)?;
                 // Prevent leak to child processes.
                 std::env::remove_var(key);
-                return Ok(Box::new(io::Cursor::new(value.into_vec())));
+                return Ok((file_name,
+                           Box::new(io::Cursor::new(value.into_vec()))));
             }
         }
 
@@ -849,7 +1085,8 @@ fn load_file<S: AsRef<str>>(file: S)
                     .map_err(|_| Error::UnsupportedSpecialPrefix)?;
                 // Prevent leak to child processes.
                 std::env::remove_var(key);
-                return Ok(Box::new(io::Cursor::new(value.into_bytes())));
+                return Ok((file_name,
+                           Box::new(io::Cursor::new(value.into_bytes()))));
             }
         }
 
@@ -857,7 +1094,9 @@ fn load_file<S: AsRef<str>>(file: S)
     }
 
     std::fs::File::open(f)
-        .map(|f| -> Box<dyn io::Read + Send + Sync> { Box::new(f) })
+        .map(|f| -> (String, Box<dyn io::Read + Send + Sync>) {
+            (file_name, Box::new(f))
+        })
         .map_err(|_| Error::MissingInput)
         .context(format!("Failed to open file {:?}", f))
 }
@@ -870,7 +1109,7 @@ fn load_file<S: AsRef<str>>(file: S)
 /// loaded from files or special designators.
 #[cfg(feature = "cli")]
 fn get_password<S: AsRef<str>>(location: S) -> Result<Password> {
-    let mut stream = load_file(location)?;
+    let mut stream = load_file(location)?.1;
     let mut pw = Vec::new();
     stream.read_to_end(&mut pw)?;
     Ok(Password::new(pw)?)
@@ -884,7 +1123,7 @@ fn get_password<S: AsRef<str>>(location: S) -> Result<Password> {
 /// loaded from files or special designators.
 #[cfg(feature = "cli")]
 fn get_password_unchecked<S: AsRef<str>>(location: S) -> Result<Password> {
-    let mut stream = load_file(location)?;
+    let mut stream = load_file(location)?.1;
     let mut pw = Vec::new();
     stream.read_to_end(&mut pw)?;
     Ok(Password::new_unchecked(pw))
@@ -930,7 +1169,7 @@ fn create_file<S: AsRef<str>>(file: S) -> Result<std::fs::File> {
 
 /// Loads the the (special) files.
 fn load_files(files: Vec<String>)
-              -> Result<Vec<Box<dyn io::Read + Send + Sync>>> {
+              -> Result<Vec<(String, Box<dyn io::Read + Send + Sync>)>> {
     files.iter().map(load_file).collect()
 }
 
@@ -980,6 +1219,29 @@ impl From<NotAfter> for std::time::SystemTime {
     }
 }
 
+/// Parses the given string depicting a ISO 8601 timestamp.
+#[derive(Debug, Clone, PartialEq, Eq, Hash)]
+struct Timestamp(DateTime<Utc>);
+
+impl std::str::FromStr for Timestamp {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> Result<Timestamp> {
+        match s {
+            // XXX: parse "-" to None once we figure out how to do that
+            // with clap.
+            "now" => Ok(Utc::now()),
+            _ => parse_iso8601(s, NaiveTime::from_hms_opt(0, 0, 0).unwrap()),
+        }.map(|t| Timestamp(t))
+    }
+}
+
+impl From<Timestamp> for std::time::SystemTime {
+    fn from(t: Timestamp) -> std::time::SystemTime {
+        t.0.into()
+    }
+}
+
 /// Parses the given string depicting a ISO 8601 timestamp.
 fn parse_iso8601(s: &str, pad_date_with: chrono::NaiveTime)
                  -> Result<DateTime<Utc>>
@@ -1055,6 +1317,10 @@ fn test_parse_iso8601() {
 /// Errors defined by the Stateless OpenPGP Command-Line Protocol.
 #[derive(thiserror::Error, Debug)]
 enum Error {
+    /// Unspecified failure.
+    #[error("Unspecified failure")]
+    UnspecifiedFailure,
+
     /// No acceptable signatures found ("sop verify").
     #[error("No acceptable signatures found")]
     NoSignature,
@@ -1139,9 +1405,17 @@ enum Error {
     #[error("Profile not supported")]
     UnsupportedProfile,
 
+    /// The primary key of a KEYS object is too weak or revoked.
+    #[error("The primary key of a KEYS object is too weak or revoked")]
+    PrimaryKeyBad,
+
+    /// The CERTS object has no matching User ID.
+    #[error("The CERTS object has no matching User ID")]
+    CertUseridNoMatch,
+
     /// An I/O operation failed.
-    #[error("I/O operation failed")]
-    IoError(#[source] io::Error),
+    #[error(transparent)]
+    IoError(#[from] io::Error),
 }
 
 impl From<crate::Error> for Error {
@@ -1149,6 +1423,7 @@ impl From<crate::Error> for Error {
         use crate::Error as E;
         use Error as CE;
         match e {
+            E::UnspecifiedFailure => CE::UnspecifiedFailure,
             E::NoSignature => CE::NoSignature,
             E::UnsupportedAsymmetricAlgo => CE::UnsupportedAsymmetricAlgo,
             E::CertCannotEncrypt => CE::CertCannotEncrypt,
@@ -1167,6 +1442,8 @@ impl From<crate::Error> for Error {
             E::IncompatibleOptions => CE::IncompatibleOptions,
             E::NotImplemented => CE::UnsupportedSubcommand,
             E::UnsupportedProfile => CE::UnsupportedProfile,
+            E::PrimaryKeyBad => CE::PrimaryKeyBad,
+            E::CertUseridNoMatch => CE::CertUseridNoMatch,
             E::IoError(e) => CE::IoError(e),
         }
     }
@@ -1176,6 +1453,7 @@ impl From<Error> for i32 {
     fn from(e: Error) -> Self {
         use Error::*;
         match e {
+            UnspecifiedFailure => 1,
             NoSignature => 3,
             UnsupportedAsymmetricAlgo => 13,
             CertCannotEncrypt => 17,
@@ -1195,6 +1473,8 @@ impl From<Error> for i32 {
             KeyCannotSign => 79,
             IncompatibleOptions => 83,
             UnsupportedProfile => 89,
+            PrimaryKeyBad => 103,
+            CertUseridNoMatch => 107,
             IoError(_) => 1,
         }
     }
